/*
 * spritemanager.cpp
 *
 *  Created on: 25-12-2012
 *      Author: crm
 */

#include "spritemanager.h"

#include <map>
#include <tinyxml.h>

#include "parser.h"
#include "resource.h"
#include "log.h"

#include "sprite.h"
#include "imagemanager.h"
#include "image.h"

/******************************************************************************/
/******************************************************************************/

using namespace Engine;

/**** Funkcje ****/
namespace Engine
	{
	namespace SpriteManager
		{
	// Funkcje pomocnicze wczytywania
		Sprite* loadImg(const std::string& name);		// Wczytanie i obsluga czystego obrazka
		Sprite* loadTxt(const std::string& name);		// Wczytanie i obsluga klasycznej wersji, z pliku tekstowego
		Sprite* loadXml(const std::string& name);		// Wczytanie i obsluga sprite w xml'u

	// Funkcje pomocnicze txt'ka
		bool txtParseAnim(Sprite *spr, const std::string& name);

	// Funkcje pomocnicze xml'a
		void xmlParseSprite(Sprite *spr, TiXmlElement *elem);
		void xmlParseAnim(Sprite::Animation *ani, TiXmlElement *elem);
		void xmlParseDir(Sprite::Direction *dir, TiXmlElement *elem);
		void xmlParseFrame(Sprite::Frame *f, TiXmlElement *elem);
		}

	/**** Zmienne ****/
	namespace SpriteManager
		{
		std::vector<Sprite*> sprites;
		std::map<std::string, Sprite*> named;
		}

	/**** Stale ****/
	namespace SpriteManager
		{
		const char XML_TAG_SPRITE[]=	"sprite";
		const char XML_TAG_IMAGE[]=		"image";
		const char XML_TAG_ANIM[]=		"anim";
		const char XML_TAG_DIR[]=		"dir";
		const char XML_TAG_FRAME[]=		"frame";
		const char XML_ATR_NAME[]=		"name";
		const char XML_ATR_RELATIVE[]=	"relative";
		const char XML_ATR_SPD[]=		"spd";
		const char XML_ATR_RET[]=		"ret";
		const char XML_ATR_X[]=			"x";
		const char XML_ATR_Y[]=			"y";
		const char XML_ATR_W[]=			"w";
		const char XML_ATR_H[]=			"h";
		const char XML_ATR_SPOTX[]=		"sx";
		const char XML_ATR_SPOTY[]=		"sy";
		const char XML_ATR_ACTX[]=		"ax";
		const char XML_ATR_ACTY[]=		"ay";
		const char XML_ATR_DIRECTION[]=	"value";
		const char XML_ATR_DEFAULT[]=	"default";

		const char XML_DIR_RIGHT[]=		"right";
		const char XML_DIR_RIGHT_DOWN[]="right-down";
		const char XML_DIR_DOWN[]=		"down";
		const char XML_DIR_LEFT_DOWN[]=	"left-down";
		const char XML_DIR_LEFT[]=		"left";
		const char XML_DIR_LEFT_UP[]=	"left-up";
		const char XML_DIR_UP[]=		"up";
		const char XML_DIR_RIGHT_UP[]=	"right-up";
		}
	}


/******************************************************************************/
/******************************************************************************/
void SpriteManager::clear()
	{
	if(sprites.size()<1)
		return;

	Log::info("SpriteManager.clear: Czyszczenie, %u sprite'ow", named.size());

	for(auto spr: sprites)
		free(spr, false);

	sprites.clear();
	named.clear();

	Log::ok("SpriteManager.clear: Wyczyszczono");
	}

/**** Wczytywanie, czesc glowna ***/
Sprite* SpriteManager::get(const std::string& name)
	{
	if(name.empty())
		return NULL;

	if(named.find(name)!=named.end())
		return named[name];

	Log::info("SpriteManager.get: Wczytywanie \"%s\"", name.c_str());

	Parser pars(name, ".");

	Sprite *spr;

	if(pars[pars.count()-1]=="xml")
		spr=loadXml(name);
	else if(pars[pars.count()-1]=="spr")
		spr=loadTxt(name);
	else
		spr=loadImg(name);

	if(!spr)
		Log::fail("SpriteManager.get: Nie udalo sie wczytac \"%s\"", name.c_str());
	else
		{
		if(spr->getAnimationCount()<1)
			Log::fail("SpriteManager.get: Sprite \"%s\" nie ma animacji", name.c_str());

		for(unsigned i=0u; i<spr->getAnimationCount(); ++i)
			{
			//if(spr->getAnim(i).getDir(DEFAULT)==NULL)
			//	Log::fail("SpriteManager.get: Sprite \"%s\" animacja %u nie ma kierunkow", name.c_str(), i);
			if(spr->getAnimation(i).getDirection(Sprite::DirectionName::DEFAULT).getFrameCount()<1)
				Log::fail("SpriteManager.get: Sprite \"%s\" animacja %u nie ma klatek w domyslnej animacji", name.c_str(), i);
			}
		}

	spr->init(name);

	Log::ok("SpriteManager.get: Wczytano \"%s\"", name.c_str());

	sprites.push_back(spr);
	named[name]=spr;

	return spr;
	}

void SpriteManager::free(const Sprite *spr, bool erase)
	{
	if(spr->refs()>0)
		{
		Log::debug("SpriteManager.free: Sprite \"%s\" ma %d referencji, nie jest kasowany", spr->getName().c_str(), spr->refs());
		return;
		}

	if(erase)
		{
		for(std::map<std::string, Sprite *>::iterator it=named.begin(); it!=named.end(); ++it)
			{
			if(it->second!=spr)
				continue;

			named.erase(it);

			break;
			}

		for(std::vector<Sprite*>::iterator it=sprites.begin(); it!=sprites.end(); ++it)
			{
			if(*it!=spr)
				continue;

			sprites.erase(it);

			break;
			}
		}

	Log::debug("SpriteManager.free: Kasowanie \"%s\"", spr->getName().c_str());

	delete spr;
	}


/******************************************************************************/
/******************************************************************************/
/**** Wczytywanie samego obrazka ****/
Sprite* SpriteManager::loadImg(const std::string& name)	// Pozwole sobie zaliczyc wczytywanie obrazka jako 'klasyczne'
	{
	Sprite *spr;
	Sprite::Animation *ani;
	Sprite::Direction *dir;
	Sprite::Frame f(0, 0, 0, 0, 0, 0, 0, 0);

	if(!Resource::find(name))
		{
		Log::fail("SpriteManager.loadImg: Nie znaleziono pliku \"%s\"", name.c_str());
		return NULL;
		}

	try
		{
		spr=new Sprite();
		}
	catch(std::bad_alloc&)
		{
		Log::fail("SpriteManager.loadImg: Nie udalo sie zajac miejsca na sprite \"%s\"", name.c_str());
		return NULL;
		}

	ImagePtr img=ImageManager::get(name.c_str());

	if(!img)
		{
		Log::fail("SpriteManager.loadImg: Nie udalo sie wczytac grafiki \"%s\"", name.c_str());
		delete spr;
		return NULL;
		}

	try
		{
		ani=new Sprite::Animation();
		dir=new Sprite::Direction();
		}
	catch(std::bad_alloc&)
		{
		Log::fail("SpriteManaget.loadImg: Nie udalo sie zajac miejsca na kierunek/animacje \"%s\"", name.c_str());
		delete spr;		// Zdaje sobie sprawe, ze przydaloby sie sprawdzic czy ani jest stworzone - wycieki i te sprawy~
		return NULL;
		}

	spr->setImage(img);

	f.x=0;		f.y=0;
	f.w=img->w;	f.h=img->h;
	f.spotx=0;	f.spoty=0;
	f.actx=0;	f.acty=0;
	dir->addFrame(f);
	ani->addDirection(Sprite::DirectionName::DEFAULT, dir);
	spr->addAnimation(ani);

	return spr;
	}


Sprite* SpriteManager::loadTxt(const std::string& name)
	{
	// Jak tu weszlo, to jest jakas nazwa.

	Parser pars("", ";");
	Sprite *spr=NULL;

	std::string namegfx=name;
	std::string nameani=name;
	char *list=Resource::load(name);
	if(list)
		{
		pars=list;
		delete [] list;

		if(pars.count()<3)
			Log::debug("SpriteManager.loadTxt: Brak wymaganych informacji o sprite \"%s\"", name.c_str());
		else
			{
			namegfx=pars[0];
			nameani=pars[1];
			}
		}

	try
		{
		spr=new Sprite();
		}
	catch(std::bad_alloc&)
		{
		Log::fail("SpriteManager.loadTxt: Nie udalo sie zajac miejsca na sprite \"%s\"", name.c_str());
		return NULL;
		}

	spr->setImage( ImageManager::get(namegfx) );

	if(!spr->getImage())
		{
		Log::fail("SpriteManager.loadTxt: Nie udalo sie wczytac obrazka \"%s\"", name.c_str());
		delete spr;
		return NULL;
		}

	if(!txtParseAnim(spr, nameani))
		{
		Log::fail("SpriteManager.loadTxt: Nie udalo sie wczytac animacji \"%s\":\"%s\"", name.c_str(), pars[1].c_str());
		delete spr;
		return NULL;
		}

	return spr;
	}

bool SpriteManager::txtParseAnim(Sprite *spr, const std::string& name)
	{
	if(name.length()<2)
		return 0;

	/*int w, h;
	w=spr->getImg()->w;
	h=spr->getImg()->h;*/

	Parser parsLine;
	Parser parsSpace;
	parsLine.setDelimiter("\n\r");
	parsSpace.setDelimiter(" \t");

	char *tmp=Resource::load(name);
	if(tmp==NULL)
		{
		Log::debug("SpriteManager.txtParseAnim: Nie znaleziono informacji o animacjach, uzywanie calego obrazka");
		Log::debug("SpriteManager.txtParseAnim: Nope, todo");

		return false;	// true;
		}

	std::string aniname="";
	Sprite::Animation *ani=NULL;
	Sprite::Direction *dir=NULL;

	parsLine=tmp;
	delete [] tmp;
	for(int i=0; i<parsLine.count(); i++)
		{
		parsSpace=parsLine[i];

		if(parsSpace.count()<1)
			continue;
		else if(parsSpace[0]=="--")
			continue;
		else if(parsSpace[0]=="anim")
			{
			if(parsSpace.count()<2)	// Wstawianie poklatkowe
				{
				Log::fail("SpriteManager.txtParseAnim: Brak wymaganych informacji (\"%s\")", parsSpace.getText().c_str());
				break;
				}

			if(ani)
				{
				spr->addAnimation(ani, aniname);

				aniname="";
				ani=NULL;
				dir=NULL;
				}

			try
				{
				ani=new Sprite::Animation();
				dir=new Sprite::Direction();
				}
			catch(std::bad_alloc &)
				{
				Log::fail("SpriteManager.txtParseAnim: Nie udalo sie zajac miejsca na animacje/kierunek \"%s\"", name.c_str());
				break;
				}

			ani->addDirection(Sprite::DirectionName::DEFAULT, dir);
			dir->setSpeed( parsSpace.valFloat(1) );
			dir->setReturn( parsSpace.val(2) );

			if(parsSpace.count()>3)
				aniname=parsSpace[3];
			}
		else if(parsSpace[0]=="frame")
			{
			int count=parsSpace.val(1);
			Sprite::Frame f( parsSpace.val(2), parsSpace.val(3), parsSpace.val(4), parsSpace.val(5) );

			if(parsSpace.count()>7)
				{
				f.spotx=parsSpace.val(6);
				f.spoty=parsSpace.val(7);
				}
			if(parsSpace.count()>9)
				{
				f.actx=parsSpace.val(8);
				f.acty=parsSpace.val(9);
				}

			for(int i=0; i<count; ++i)
				dir->addFrame(f);
			}
		//else if(parsSpace[0]=="anim")
		else if(parsSpace.isVal(0) && parsSpace.count()>6)			// Animacje 'na szybko'	// Proste zalozenie, jesli pierwsza wartosc jest liczba to dodawana jest animacja
			{
			// 0  1  2  3  4  5  6
			// c  x  y  w  h sx sy
			// 5  0 24 12 24  7 24

			//printf("%s", parsSpace.getText().c_str());

			if(ani)
				{
				spr->addAnimation(ani, aniname);

				aniname="";
				ani=NULL;
				dir=NULL;
				}

			try
				{
				ani=new Sprite::Animation();
				dir=new Sprite::Direction();
				}
			catch(std::bad_alloc&)
				{
				Log::fail("SpriteManager.txtParseAnims: Nie udalo sie zajac miejsca na animacje/kierunek \"%s\"", name.c_str());
				break;
				}

			ani->addDirection(Sprite::DirectionName::DEFAULT, dir);

			if(parsSpace.count()<7)
				{
				Log::fail("Sprite.loadAnims: Brak wymaganych informacji (\"%s\")", parsSpace.getText().c_str());
				break;
				}

			int count=parsSpace.val(0);
			Sprite::Frame f( parsSpace.val(1), parsSpace.val(2), parsSpace.val(3), parsSpace.val(4), parsSpace.val(5), parsSpace.val(6) );

			for(int i=0; i<count; ++i)
				{
				dir->addFrame(f);
				f.x+=f.w;
				}
			}
		}

	if(ani)
		spr->addAnimation(ani, aniname);

// Sprawdz czy wczytala sie przynajmniej jedna animacja
	if(spr->getAnimationCount()<1)
		return false;
	for(unsigned i=0u; i<spr->getAnimationCount(); ++i)
		{
		if(spr->getAnimation(i).getDirection(Sprite::DirectionName::DEFAULT).getFrameCount()<1)
			// Dodaj pusta klatke? Chyba sie nie bedzie dalo.
			return false;
		}

	return true;
	}


/**** Wczytywanie xml ****/
Sprite* SpriteManager::loadXml(const std::string& name)
	{
	char *data;

	Sprite *spr;
	TiXmlDocument xml(""); // <- WAZNE! NIE mozna podsylac NULLa do tego konstruktora.

	data=Resource::load(name);
	if(!data)
		{
		Log::fail("SpriteManager.loadXml: Nie udalo sie wczytac xml'a \"%s\"", name.c_str());
		return NULL;
		}

	xml.Parse(data);
	delete [] data;

	TiXmlNode *node=xml.FirstChild(XML_TAG_SPRITE);

	if(!node)
		{
		Log::fail("SpriteManager.loadXml: Nie znaleziono tagu \"sprite\" w pliku \"%s\"", name.c_str());
		return NULL;
		}

	const char *sname;
	sname=node->ToElement()->Attribute(XML_ATR_NAME);
	sname=(sname)?sname:name.c_str();

	try
		{
		spr=new Sprite();
		}
	catch(std::bad_alloc&)
		{
		Log::fail("SpriteManager.loadXml: Nie udalo sie zajac miejsca na sprite \"%s\"", name.c_str());
		return NULL;
		}

	// Sprajt
	TiXmlElement *img=node->ToElement()->FirstChildElement(XML_TAG_IMAGE);

	if(img)
		{
		if(!img->Attribute(XML_ATR_RELATIVE) || name.length()<1)
			spr->setImage( ImageManager::get(img->GetText()) );
		else
			{
			Parser p(name, "/\\:");

			p.del(p.count()-1);
			p.ins(p.count(), img->GetText());

			Log::debug("SpriteManager.loadXml: Relatywna sciezka do rysunku: \"%s\" w \"%s\"", p.getText().c_str(), name.c_str());

			spr->setImage( ImageManager::get(p.getText()) );
			}
		}
	else
		Log::fail("SpriteManager.xmlParseSprite: Nie ustawiono obrazka");

	xmlParseSprite(spr, node->ToElement());

	// Walidacja
	if(spr->getAnimationCount()<1)
		{
		Log::debug("SpriteManager.loadXml: Nie dodano zadnej animacji, dodawanie calego obrazka");

		Sprite::Animation *a=new Sprite::Animation();
		Sprite::Direction *d=new Sprite::Direction();

		d->setReturn(0u);
		d->setSpeed(0.0f);
		d->addFrame( Sprite::Frame(0, 0, spr->getImage()->w, spr->getImage()->h, 0, 0, 0, 0) );

		a->addDirection(Sprite::DirectionName::DEFAULT, d);
		spr->addAnimation(a);
		}

	return spr;
	}

void SpriteManager::xmlParseSprite(Sprite *spr, TiXmlElement *elem)
	{
	// Wczytanie animacji
	TiXmlNode *node=elem->FirstChild(XML_TAG_ANIM);
	while(node)
		{
		const char *name;

		TiXmlElement *e=node->ToElement();

		name=e->Attribute(XML_ATR_NAME);

		//Log::debug("SpriteManager.xmlParseSprite: Animacja, spd %f ret %d", ani.getSpd(), ani.getRet());

		Sprite::Animation *ani;

		try {
			ani=new Sprite::Animation();
			}
		catch(std::bad_alloc& e)
			{
			continue;
			}

		//Log::debug("SpriteManaget.xmlParseSprite: anim, name \"%s\"", name);
		xmlParseAnim(ani, e);

		/*unsigned aaa=*/spr->addAnimation(ani, (name)?name:"");
		//Log::debug("SpriteManaget.xmlParseSprite: anim, name \"%s\" %u", name, aaa);

		// Dalsza iteracja
		node=elem->IterateChildren(XML_TAG_ANIM, node);	// Z jakiegos powodu bylo tutaj: node=elem->IterateChildren(XML_TAG_FRAME, node);
		}
	}

void SpriteManager::xmlParseAnim(Sprite::Animation *ani, TiXmlElement *elem)
	{
	TiXmlNode *node;

	const char *dirstr;
	double spd;
	int ret;

	Sprite::DirectionName dirv;

	node=elem->FirstChild(XML_TAG_DIR);
	while(node)
		{
		TiXmlElement *e=node->ToElement();

		spd=1.0f;
		ret=-1;

		dirstr=e->Attribute(XML_ATR_DIRECTION);
		e->Attribute(XML_ATR_SPD, &spd);
		e->Attribute(XML_ATR_RET, &ret);

		if(!dirstr)										dirv=Sprite::DirectionName::DEFAULT;
		else if(strcmp(dirstr, XML_DIR_RIGHT)==0)		dirv=Sprite::DirectionName::RIGHT;
		else if(strcmp(dirstr, XML_DIR_RIGHT_DOWN)==0)	dirv=Sprite::DirectionName::RIGHT_DOWN;
		else if(strcmp(dirstr, XML_DIR_DOWN)==0)		dirv=Sprite::DirectionName::DOWN;
		else if(strcmp(dirstr, XML_DIR_LEFT_DOWN)==0)	dirv=Sprite::DirectionName::LEFT_DOWN;
		else if(strcmp(dirstr, XML_DIR_LEFT)==0)		dirv=Sprite::DirectionName::LEFT;
		else if(strcmp(dirstr, XML_DIR_LEFT_UP)==0)		dirv=Sprite::DirectionName::LEFT_UP;
		else if(strcmp(dirstr, XML_DIR_UP)==0)			dirv=Sprite::DirectionName::UP;
		else if(strcmp(dirstr, XML_DIR_RIGHT_UP)==0)	dirv=Sprite::DirectionName::RIGHT_UP;
		else
			{
			dirv=Sprite::DirectionName::DEFAULT;

			Log::fail("SpriteManager.xmlParseAnim: Nieprawidlowy kierunek \"%s\"", dirstr);
			}

		Sprite::Direction *dir;

		try
			{
			dir=new Sprite::Direction();
			}
		catch(std::bad_alloc&)
			{
			continue;
			}

		dir->setSpeed(spd);
		dir->setReturn(ret);

		//Log::debug("SpriteManaget.xmlParseSprite: dir, spd %f ret %d dir %d (%d)", spd, ret, dirv, (Directions)dirv);
		xmlParseDir(dir, e);

		if(ret<0)
			dir->setReturn(dir->getFrameCount()-1);

		ani->addDirection(dirv, dir);

		node=elem->IterateChildren(XML_TAG_DIR, node);
		}

	node=elem->FirstChild(XML_TAG_FRAME);
	if(node)
		{
		Sprite::Direction *dir;

		try
			{
			dir=new Sprite::Direction();
			}
		catch(std::bad_alloc&)
			{
			return;
			}

		//Log::debug("SpriteManaget.xmlParseSprite: dir, spd %f ret %d dir %d (%d)", spd, ret, dirv, (Directions)dirv);
		xmlParseDir(dir, elem);

		dir->setSpeed(0);
		dir->setReturn(dir->getFrameCount()-1);

		ani->addDirection(Sprite::DirectionName::DEFAULT, dir);
		}
	}

void SpriteManager::xmlParseDir(Sprite::Direction *dir, TiXmlElement *elem)
	{
	TiXmlNode *node;

	node=elem->FirstChild(XML_TAG_FRAME);
	while(node)
		{
		TiXmlElement *e=node->ToElement();

		Sprite::Frame f(0, 0, 0, 0, 0, 0, 0, 0);

		xmlParseFrame(&f, e);

		//Log::debug("SpriteManaget.xmlParseSprite: frame");
		dir->addFrame(f);

		node=elem->IterateChildren(XML_TAG_FRAME, node);
		}
	}

void SpriteManager::xmlParseFrame(Sprite::Frame *f, TiXmlElement *elem)
	{
	if(!elem->Attribute(XML_ATR_X,		&f->x)) 	f->x=0;
	if(!elem->Attribute(XML_ATR_Y,		&f->y)) 	f->y=0;
	if(!elem->Attribute(XML_ATR_W,		&f->w)) 	f->w=0;
	if(!elem->Attribute(XML_ATR_H,		&f->h)) 	f->h=0;
	if(!elem->Attribute(XML_ATR_SPOTX,	&f->spotx)) f->spotx=0;
	if(!elem->Attribute(XML_ATR_SPOTY,	&f->spoty)) f->spoty=0;
	if(!elem->Attribute(XML_ATR_ACTX,	&f->actx))  f->actx=0;
	if(!elem->Attribute(XML_ATR_ACTY,	&f->acty))  f->acty=0;
	}
