/*
 * objecttemplate.cpp
 *
 *  Created on: 27 kwi 2016
 *      Author: crm
 */

#include "objecttemplate.h"

#include <tinyxml.h>

#include "log.h"
#include "resource.h"
#include "parser.h"

#include "../consts.h"
//#include "object.h"

using namespace Game;

bool ObjectTemplate::init(const std::string& path)
	{
	const char XML_TAG_OBJECT[]=	"object";
	const char XML_TAG_SPRITE[]=	"sprite";
	const char XML_TAG_BOUNDARY[]=	"boundary";
	const char XML_TAG_COLLIDERS[]=	"colliders";
	const char XML_TAG_BOX[]=		"box";
	const char XML_ATR_NAME[]=		"name";
	const char XML_ATR_RELATIVE[]=	"relative";
	const char XML_ATR_MAX_SPD[]=	"speed";
	const char XML_ATR_X[]=			"x";
	const char XML_ATR_Y[]=			"y";
	const char XML_ATR_W[]=			"w";
	const char XML_ATR_H[]=			"h";
	const char XML_ATR_FRICTION[]=	"friction";

	char *data;

	TiXmlDocument xml("");

	bool bboxset=false;

	Log::info("ObjectTemplate.init: Wczytywanie pliku \"%s\"", path.c_str());

	data=Resource::load(path);
	if(!data)
		{
		Log::fail("ObjectTemplate.init: Nie udalo sie wczytac xml'a \"%s\"", path.c_str());
		return false;
		}

	xml.Parse(data);
	delete [] data;

	// Obiekt
	TiXmlNode *xobj=xml.FirstChild(XML_TAG_OBJECT);

	if(!xobj)
		{
		Log::fail("ObjectTemplate.init: Nie znaleziono tagu \"object\" w \"%s\"", path.c_str());
		return false;
		}

	const char *xaname;
	xaname=xobj->ToElement()->Attribute(XML_ATR_NAME);

	if(!xaname)
		{
		Log::fail("ObjectTemplate.init: Obiekt nie ma nazwy w \"%s\"", path.c_str());
		return false;
		}

	name=xaname;

	// Szybkosc
	double aspeed;
	if(!xobj->ToElement()->Attribute(XML_ATR_MAX_SPD, &aspeed))
		{
		Log::info("ObjectTemplate.init: Obiekt nie ma maksymalnej szybkosci w \"%s\"", path.c_str());
		maxSpeed=1.0f*METERS_TO_PIXELS;
		}
	else
		{
		maxSpeed=aspeed*METERS_TO_PIXELS;
		Log::info("ObjectTemplate.init: Maksymalna szybkosc: %.2fm/s w \"%s\"", aspeed, path.c_str());
		}

	Log::info("ObjectTemplate.init: Obiekt \"%s\" w \"%s\"", name.c_str(), path.c_str());

	// Sprite
	TiXmlElement *xspr=xobj->ToElement()->FirstChildElement(XML_TAG_SPRITE);

	if(!xspr)
		{
		Log::fail("ObjectTemplate.init: Nie znaleziono tagu \"sprite\" w \"%s\"", path.c_str());
		return false;
		}

	if(!xspr->Attribute(XML_ATR_RELATIVE))
		sprite.setSprite(xspr->GetText());
	else
		{
		Parser p(path, "/\\:");

		p.del(p.count()-1);
		p.ins(p.count(), xspr->GetText());

		Log::info("ObjectTemplate.init: Relatywna sciezka do sprite: \"%s\" w \"%s\"", p.getText().c_str(), path.c_str());

		sprite.setSprite(p.getText());
		}

	// Bounding Box
	TiXmlElement *xbbox=xobj->ToElement()->FirstChildElement(XML_TAG_BOUNDARY);

	if(!xbbox)
		{
		Log::info("ObjectTemplate.init: Nie znaleziono tagu \"%s\" w \"%s\", bounding box zostanie ustawiony automatycznie", XML_TAG_BOUNDARY, path.c_str());
		}
	else
		{
		double x, y, w, h;

		bboxset=true;

		if(!xbbox->Attribute(XML_ATR_X, &x))
			{
			Log::fail("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_X, XML_TAG_BOUNDARY, path.c_str());
			bboxset=false;
			}
		if(!xbbox->Attribute(XML_ATR_Y, &y))
			{
			Log::fail("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_Y, XML_TAG_BOUNDARY, path.c_str());
			bboxset=false;
			}
		if(!xbbox->Attribute(XML_ATR_W, &w))
			{
			Log::fail("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_W, XML_TAG_BOUNDARY, path.c_str());
			bboxset=false;
			}
		if(!xbbox->Attribute(XML_ATR_H, &h))
			{
			Log::fail("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_H, XML_TAG_BOUNDARY, path.c_str());
			bboxset=false;
			}

		if(bboxset)
			{
			boundingbox=Box(AVector(x, y, 0, 0), AVector(w, h, 0, 0));
			}
		}

	// Collidery
	TiXmlElement *xcol=xobj->ToElement()->FirstChildElement(XML_TAG_COLLIDERS);

	if(!xcol)
		{
		Log::info("ObjectTemplate.init: Nie znaleziono tagu \"%s\" w \"%s\", obiekt nie bedzie mial colliderow", XML_TAG_COLLIDERS, path.c_str());
		}
	else
		{
		TiXmlElement *xbox=xcol->ToElement()->FirstChildElement(XML_TAG_BOX);

		if(!xbox)
			{
			Log::fail("ObjectTemplate.init: Nie znaleziono tagu \"%s\" w tagu \"%s\" w \"%s\"", XML_TAG_BOX, XML_TAG_COLLIDERS, path.c_str());

			return false;
			}

		while(xbox)
			{
			double x, y, w, h, friction;

			if(!xbox->Attribute(XML_ATR_X, &x))
				{
				Log::fail("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_X, XML_TAG_BOX, path.c_str());
				break;
				}

			if(!xbox->Attribute(XML_ATR_Y, &y))
				{
				Log::fail("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_Y, XML_TAG_BOX, path.c_str());
				break;
				}

			if(!xbox->Attribute(XML_ATR_W, &w))
				{
				Log::fail("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_W, XML_TAG_BOX, path.c_str());
				break;
				}

			if(!xbox->Attribute(XML_ATR_H, &h))
				{
				Log::fail("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\"", XML_ATR_H, XML_TAG_BOX, path.c_str());
				break;
				}

			if(!xbox->Attribute(XML_ATR_FRICTION, &friction))
				{
				Log::info("ObjectTemplate.init: Nie znaleziono atrybutu \"%s\" tagu \"%s\" w \"%s\", ustawianie na 1.0", XML_ATR_FRICTION, XML_TAG_BOX, path.c_str());
				friction=1.0;
				}

			colliders.push_back(Collider(AVector(x, y, 0, 0), AVector(w, h, 0, 0), friction));

			xbox=xbox->NextSiblingElement(XML_TAG_BOX);
			}
		}

	// Wyznaczanie BB
	Log::info("ObjectTemplate.init: Wczytano %u colliderow w \"%s\"", colliders.size(), path.c_str());

	if(bboxset)
		{
		Log::info("ObjectTemplate.init: Bounding Box ustawiony manualnie w \"%s\"", path.c_str());
		}
	else
		{
		Log::info("ObjectTemplate.init: Bounding Box ustawiany automatycznie w \"%s\"", path.c_str());

		boundingbox=Box(AVector(), AVector());

		if(colliders.size()>0u)
			boundingbox=colliders[0];

		for(auto c: colliders)
			{
			if(c.pos.x<boundingbox.pos.x)
				{
				boundingbox.size.x+=boundingbox.pos.x-c.pos.x;
				boundingbox.pos.x=c.pos.x;
				}
			if(c.pos.y<boundingbox.pos.y)
				{
				boundingbox.size.y+=boundingbox.pos.y-c.pos.y;
				boundingbox.pos.y=c.pos.y;
				}
			if(c.pos.x+c.size.x>boundingbox.pos.x+boundingbox.size.x)
				{
				boundingbox.size.x+=c.pos.x+c.size.x-(boundingbox.pos.x+boundingbox.size.x);
				}
			if(c.pos.y+c.size.y>boundingbox.pos.y+boundingbox.size.y)
				{
				boundingbox.size.y+=c.pos.y+c.size.y-(boundingbox.pos.y+boundingbox.size.y);
				}
			}
		}

	//Log::debug("ObjectTemplate.init: BoundingBox: %f,%f; %f,%f w \"%s\"", boundingbox.pos.x, boundingbox.pos.y, boundingbox.size.x, boundingbox.size.y, path.c_str());
	Log::ok("ObjectTemplate.init: Wczytano \"%s\" z pliku \"%s\"", name.c_str(), path.c_str());

	this->path=path;

	return true;
	}

void ObjectTemplate::clear()
	{
	sprite.clear();
	}
